@@ -20,6 +20,9 @@ from netplan.cli.commands.generate import NetplanGenerate
 from netplan.cli.commands.ip import NetplanIp
 from netplan.cli.commands.migrate import NetplanMigrate
 from netplan.cli.commands.try_command import NetplanTry
+from netplan.cli.commands.info import NetplanInfo
+from netplan.cli.commands.set import NetplanSet
+from netplan.cli.commands.get import NetplanGet
 
 __all__ = [
     'NetplanApply',
@@ -27,4 +30,7 @@ __all__ = [
     'NetplanIp',
     'NetplanMigrate',
     'NetplanTry',
+    'NetplanInfo',
+    'NetplanSet',
+    'NetplanGet',
 ]
