@@ -19,7 +19,7 @@
 import os
 import textwrap
 
-from .base import TestBase, ND_DHCP4, ND_DHCP6, ND_DHCPYES
+from .base import TestBase, ND_DHCP4, ND_DHCP6, ND_DHCPYES, ND_EMPTY
 
 
 class TestNetworkd(TestBase):
@@ -104,13 +104,26 @@ ConfigureWithoutCarrier=yes
             'bond0.network': '''[Match]
 Name=bond0
 
+[Link]
+MTUBytes=9000
+
 [Network]
 LinkLocalAddressing=ipv6
 ConfigureWithoutCarrier=yes
 VLAN=bond0.108
 ''',
             'eth1.link': '[Match]\nOriginalName=eth1\n\n[Link]\nWakeOnLan=off\nMTUBytes=9000\n',
-            'eth1.network': '[Match]\nName=eth1\n\n[Network]\nLinkLocalAddressing=no\nIPv6MTUBytes=2000\nBond=bond0\n'
+            'eth1.network': '''[Match]
+Name=eth1
+
+[Link]
+MTUBytes=9000
+
+[Network]
+LinkLocalAddressing=no
+IPv6MTUBytes=2000
+Bond=bond0
+'''
         })
         self.assert_networkd_udev(None)
 
@@ -278,12 +291,110 @@ RouteMetric=100
 UseMTU=true
 '''})
 
+    def test_eth_address_option_lifetime_zero(self):
+        self.generate('''network:
+  version: 2
+  ethernets:
+    engreen:
+      addresses:
+        - 192.168.14.2/24:
+            lifetime: 0
+        - 2001:FFfe::1/64''')
+
+        self.assert_networkd({'engreen.network': '''[Match]
+Name=engreen
+
+[Network]
+LinkLocalAddressing=ipv6
+Address=2001:FFfe::1/64
+
+[Address]
+Address=192.168.14.2/24
+PreferredLifetime=0
+'''})
+
+    def test_eth_address_option_lifetime_forever(self):
+        self.generate('''network:
+  version: 2
+  ethernets:
+    engreen:
+      addresses:
+        - 192.168.14.2/24:
+            lifetime: forever
+        - 2001:FFfe::1/64''')
+
+        self.assert_networkd({'engreen.network': '''[Match]
+Name=engreen
+
+[Network]
+LinkLocalAddressing=ipv6
+Address=2001:FFfe::1/64
+
+[Address]
+Address=192.168.14.2/24
+PreferredLifetime=forever
+'''})
+
+    def test_eth_address_option_label(self):
+        self.generate('''network:
+  version: 2
+  ethernets:
+    engreen:
+      addresses:
+        - 192.168.14.2/24:
+            label: test-label
+        - 2001:FFfe::1/64''')
+
+        self.assert_networkd({'engreen.network': '''[Match]
+Name=engreen
+
+[Network]
+LinkLocalAddressing=ipv6
+Address=2001:FFfe::1/64
+
+[Address]
+Address=192.168.14.2/24
+Label=test-label
+'''})
+
+    def test_eth_address_option_multi_pass(self):
+        self.generate('''network:
+  version: 2
+  bridges:
+    br0:
+      interfaces: [engreen]
+  ethernets:
+    engreen:
+      addresses:
+        - 192.168.14.2/24:
+            label: test-label
+        - 2001:FFfe::1/64:
+            label: ip6''')
+
+        self.assert_networkd({
+            'engreen.network': '''[Match]
+Name=engreen
+
+[Network]
+LinkLocalAddressing=no
+Bridge=br0
+
+[Address]
+Address=192.168.14.2/24
+Label=test-label
+
+[Address]
+Address=2001:FFfe::1/64
+Label=ip6
+''',
+            'br0.network': ND_EMPTY % ('br0', 'ipv6'),
+            'br0.netdev': '[NetDev]\nName=br0\nKind=bridge\n'})
+
     def test_dhcp_critical_true(self):
         self.generate('''network:
   version: 2
   ethernets:
     engreen:
-      dhcp4: yes
       critical: yes
 ''')
 
@@ -291,13 +402,10 @@ UseMTU=true
 Name=engreen
 
 [Network]
-DHCP=ipv4
 LinkLocalAddressing=ipv6
 
 [DHCP]
 CriticalConnection=true
-RouteMetric=100
-UseMTU=true
 '''})
 
     def test_dhcp_identifier_mac(self):
@@ -500,6 +608,56 @@ Name=engreen
 DHCP=yes
 LinkLocalAddressing=no
 
+[DHCP]
+RouteMetric=100
+UseMTU=true
+'''})
+
+    def test_ip6_addr_gen_mode(self):
+        self.generate('''network:
+  version: 2
+  renderer: networkd
+  ethernets:
+    enblue:
+      dhcp6: yes
+      ipv6-address-generation: eui64''')
+        self.assert_networkd({'enblue.network': '''[Match]\nName=enblue\n
+[Network]
+DHCP=ipv6
+LinkLocalAddressing=ipv6
+
+[DHCP]
+RouteMetric=100
+UseMTU=true
+'''})
+
+    def test_ip6_addr_gen_token(self):
+        self.generate('''network:
+  version: 2
+  renderer: networkd
+  ethernets:
+    engreen:
+      dhcp6: yes
+      ipv6-address-token: ::2
+    enblue:
+      dhcp6: yes
+      ipv6-address-token: "::2"''')
+        self.assert_networkd({'engreen.network': '''[Match]\nName=engreen\n
+[Network]
+DHCP=ipv6
+LinkLocalAddressing=ipv6
+IPv6Token=static:::2
+
+[DHCP]
+RouteMetric=100
+UseMTU=true
+''',
+                              'enblue.network': '''[Match]\nName=enblue\n
+[Network]
+DHCP=ipv6
+LinkLocalAddressing=ipv6
+IPv6Token=static:::2
+
 [DHCP]
 RouteMetric=100
 UseMTU=true
@@ -713,6 +871,92 @@ method=auto
 
 [ipv6]
 method=auto
+'''})
+
+    def test_ip6_addr_gen_mode(self):
+        self.generate('''network:
+  version: 2
+  renderer: NetworkManager
+  ethernets:
+    engreen:
+      dhcp6: yes
+      ipv6-address-generation: stable-privacy
+    enblue:
+      dhcp6: yes
+      ipv6-address-generation: eui64''')
+        self.assert_nm({'engreen': '''[connection]
+id=netplan-engreen
+type=ethernet
+interface-name=engreen
+
+[ethernet]
+wake-on-lan=0
+
+[ipv4]
+method=link-local
+
+[ipv6]
+method=auto
+addr-gen-mode=1
+''',
+                        'enblue': '''[connection]
+id=netplan-enblue
+type=ethernet
+interface-name=enblue
+
+[ethernet]
+wake-on-lan=0
+
+[ipv4]
+method=link-local
+
+[ipv6]
+method=auto
+addr-gen-mode=0
+'''})
+
+    def test_ip6_addr_gen_token(self):
+        self.generate('''network:
+  version: 2
+  renderer: NetworkManager
+  ethernets:
+    engreen:
+      dhcp6: yes
+      ipv6-address-token: ::2
+    enblue:
+      dhcp6: yes
+      ipv6-address-token: "::2"''')
+        self.assert_nm({'engreen': '''[connection]
+id=netplan-engreen
+type=ethernet
+interface-name=engreen
+
+[ethernet]
+wake-on-lan=0
+
+[ipv4]
+method=link-local
+
+[ipv6]
+method=auto
+addr-gen-mode=0
+token=::2
+''',
+                        'enblue': '''[connection]
+id=netplan-enblue
+type=ethernet
+interface-name=enblue
+
+[ethernet]
+wake-on-lan=0
+
+[ipv4]
+method=link-local
+
+[ipv6]
+method=auto
+addr-gen-mode=0
+token=::2
 '''})
 
     def test_eth_manual_addresses(self):
