@@ -20,7 +20,7 @@ import os
 import re
 import unittest
 
-from .base import TestBase
+from .base import TestBase, ND_VLAN, ND_EMPTY, ND_WITHIP, ND_DHCP6_WOCARRIER
 
 
 class TestNetworkd(TestBase):
@@ -51,20 +51,8 @@ VLAN=enblue
 VLAN=enred
 VLAN=engreen
 ''',
-                              'enblue.netdev': '''[NetDev]
-Name=enblue
-Kind=vlan
-
-[VLAN]
-Id=1
-''',
-                              'engreen.netdev': '''[NetDev]
-Name=engreen
-Kind=vlan
-
-[VLAN]
-Id=2
-''',
+                              'enblue.netdev': ND_VLAN % ('enblue', 1),
+                              'engreen.netdev': ND_VLAN % ('engreen', 2),
                               'enred.netdev': '''[NetDev]
 Name=enred
 MACAddress=aa:bb:cc:dd:ee:11
@@ -73,38 +61,86 @@ Kind=vlan
 [VLAN]
 Id=3
 ''',
-                              'enblue.network': '''[Match]
-Name=enblue
+                              'enblue.network': ND_WITHIP % ('enblue', '1.2.3.4/24'),
+                              'enred.network': ND_EMPTY % ('enred', 'ipv6'),
+                              'engreen.network': (ND_DHCP6_WOCARRIER % 'engreen')})
 
-[Network]
-LinkLocalAddressing=ipv6
-Address=1.2.3.4/24
-ConfigureWithoutCarrier=yes
-''',
-                              'enred.network': '''[Match]
-Name=enred
-
-[Network]
-LinkLocalAddressing=ipv6
-ConfigureWithoutCarrier=yes
-''',
-                              'engreen.network': '''[Match]
-Name=engreen
-
-[Network]
-DHCP=ipv6
-LinkLocalAddressing=ipv6
-ConfigureWithoutCarrier=yes
-
-[DHCP]
-RouteMetric=100
-UseMTU=true
-'''})
         self.assert_nm(None, '''[keyfile]
 # devices managed by networkd
 unmanaged-devices+=interface-name:en1,interface-name:enblue,interface-name:enred,interface-name:engreen,''')
         self.assert_nm_udev(None)
 
+    def test_vlan_sriov(self):
+        # we need to make sure renderer: sriov vlans are not saved as part of
+        # the NM/networkd config
+        self.generate('''network:
+  version: 2
+  ethernets:
+    en1: {}
+  vlans:
+    enblue:
+      id: 1
+      link: en1
+      renderer: sriov
+    engreen: {id: 2, link: en1, dhcp6: true}''')
+
+        self.assert_networkd({'en1.network': '''[Match]
+Name=en1
+
+[Network]
+LinkLocalAddressing=ipv6
+VLAN=engreen
+''',
+                              'engreen.netdev': ND_VLAN % ('engreen', 2),
+                              'engreen.network': (ND_DHCP6_WOCARRIER % 'engreen')})
+
+        self.assert_nm(None, '''[keyfile]
+# devices managed by networkd
+unmanaged-devices+=interface-name:en1,interface-name:enblue,interface-name:engreen,''')
+        self.assert_nm_udev(None)
+
+    # see LP: #1888726
+    def test_vlan_parent_match(self):
+        self.generate('''network:
+  version: 2
+  renderer: networkd
+  ethernets:
+    lan:
+      match: {macaddress: "11:22:33:44:55:66"}
+      set-name: lan
+      mtu: 9000
+  vlans:
+    vlan20: {id: 20, link: lan}''')
+
+        self.assert_networkd({'lan.network': '''[Match]
+MACAddress=11:22:33:44:55:66
+Name=lan
+Type=!vlan bond bridge
+
+[Link]
+MTUBytes=9000
+
+[Network]
+LinkLocalAddressing=ipv6
+VLAN=vlan20
+''',
+                              'lan.link': '''[Match]
+MACAddress=11:22:33:44:55:66
+Type=!vlan bond bridge
+
+[Link]
+Name=lan
+WakeOnLan=off
+MTUBytes=9000
+''',
+                              'vlan20.network': ND_EMPTY % ('vlan20', 'ipv6'),
+                              'vlan20.netdev': ND_VLAN % ('vlan20', 20)})
+
+        self.assert_nm(None, '''[keyfile]
+# devices managed by networkd
+unmanaged-devices+=mac:11:22:33:44:55:66,interface-name:vlan20,''')
+        self.assert_nm_udev(None)
+
 
 class TestNetworkManager(TestBase):
 
@@ -221,3 +257,51 @@ method=auto
 method=ignore
 ''' % uuid})
         self.assert_nm_udev(None)
+
+    def test_vlan_sriov(self):
+        # we need to make sure renderer: sriov vlans are not saved as part of
+        # the NM/networkd config
+        self.generate('''network:
+  version: 2
+  renderer: NetworkManager
+  ethernets:
+    en1: {}
+  vlans:
+    enblue:
+      id: 1
+      link: en1
+      addresses: [1.2.3.4/24]
+      renderer: sriov
+    engreen: {id: 2, link: en1, dhcp6: true}''')
+
+        self.assert_networkd({})
+        self.assert_nm({'en1': '''[connection]
+id=netplan-en1
+type=ethernet
+interface-name=en1
+
+[ethernet]
+wake-on-lan=0
+
+[ipv4]
+method=link-local
+
+[ipv6]
+method=ignore
+''',
+                        'engreen': '''[connection]
+id=netplan-engreen
+type=vlan
+interface-name=engreen
+
+[vlan]
+id=2
+parent=en1
+
+[ipv4]
+method=link-local
+
+[ipv6]
+method=auto
+'''})
+        self.assert_nm_udev(None)
