@@ -486,6 +486,38 @@ class TestNetworkd(IntegrationTestsBase, _CommonTests):
         with open('/sys/class/net/mybond/bonding/arp_ip_target') as f:
             self.assertEqual(f.read().strip(), '192.168.5.1')
 
+    def test_bond_arp_targets_many_lp1829264(self):
+        self.setup_eth(None)
+        self.start_dnsmasq(None, self.dev_e2_ap)
+        self.addCleanup(subprocess.call, ['ip', 'link', 'delete', 'mybond'], stderr=subprocess.DEVNULL)
+        with open(self.config, 'w') as f:
+            f.write('''network:
+  renderer: %(r)s
+  ethernets:
+    ethbn:
+      match: {name: %(ec)s}
+    %(e2c)s: {}
+  bonds:
+    mybond:
+      interfaces: [ethbn]
+      parameters:
+        mode: balance-xor
+        arp-interval: 50000
+        arp-ip-targets: [ 192.168.5.1, 192.168.5.34 ]
+      dhcp4: yes''' % {'r': self.backend, 'ec': self.dev_e_client, 'e2c': self.dev_e2_client})
+        self.generate_and_settle()
+        self.assert_iface_up(self.dev_e_client,
+                             ['master mybond'],
+                             ['inet '])
+        self.assert_iface_up('mybond',
+                             ['inet 192.168.5.[0-9]+/24'])
+        with open('/sys/class/net/mybond/bonding/slaves') as f:
+            self.assertEqual(f.read().strip(), self.dev_e_client)
+        with open('/sys/class/net/mybond/bonding/arp_ip_target') as f:
+            result = f.read().strip()
+            self.assertIn('192.168.5.1', result)
+            self.assertIn('192.168.5.34', result)
+
     def test_bond_arp_all_targets(self):
         self.setup_eth(None)
         self.start_dnsmasq(None, self.dev_e2_ap)
